"""Base class for the feed manager. This allows managing feeds and their entries throughout their life-cycle."""

from __future__ import annotations

from datetime import datetime
import logging
from typing import Awaitable, Callable

from .consts import UPDATE_OK, UPDATE_OK_NO_DATA
from .feed import QuakeMLFeed
from .feed_entry import FeedEntry
from .status_update import StatusUpdate

_LOGGER = logging.getLogger(__name__)


class QuakeMLFeedManagerBase:
    """Generic Feed manager."""

    def __init__(
        self,
        feed: QuakeMLFeed,
        generate_async_callback: Callable[[str], Awaitable[None]] | None = None,
        update_async_callback: Callable[[str], Awaitable[None]] | None = None,
        remove_async_callback: Callable[[str], Awaitable[None]] | None = None,
        status_async_callback: Callable[[StatusUpdate], Awaitable[None]] | None = None,
    ):
        """Initialise feed manager."""
        self._feed: QuakeMLFeed = feed
        self.feed_entries: dict = {}
        self._managed_external_ids: set = set()
        self._last_update: datetime | None = None
        self._last_update_successful: datetime | None = None
        self._generate_async_callback: Callable[[str], Awaitable[None]] = (
            generate_async_callback
        )
        self._update_async_callback: Callable[[str], Awaitable[None]] = (
            update_async_callback
        )
        self._remove_async_callback: Callable[[str], Awaitable[None]] = (
            remove_async_callback
        )
        self._status_async_callback: Callable[[StatusUpdate], Awaitable[None]] = (
            status_async_callback
        )

    def __repr__(self):
        """Return string representation of this feed."""
        return f"<{self.__class__.__name__}(feed={self._feed})>"

    async def update(self):
        """Update the feed and then update connected entities."""
        status, feed_entries = await self._feed.update()
        # Record current time of update.
        self._last_update = datetime.now()
        count_created: int = 0
        count_updated: int = 0
        count_removed: int = 0
        await self._store_feed_entries(status, feed_entries)
        if status == UPDATE_OK:
            _LOGGER.debug("Data retrieved %s", feed_entries)
            # Record current time of update.
            self._last_update_successful = self._last_update
            # For entity management the external ids from the feed are used.
            feed_external_ids = {entry.external_id for entry in feed_entries}
            count_removed = await self._update_feed_remove_entries(feed_external_ids)
            count_updated = await self._update_feed_update_entries(feed_external_ids)
            count_created = await self._update_feed_create_entries(feed_external_ids)
        elif status == UPDATE_OK_NO_DATA:
            _LOGGER.debug("Update successful, but no data received from %s", self._feed)
            # Record current time of update.
            self._last_update_successful = self._last_update
        else:
            _LOGGER.warning(
                "Update not successful, no data received from %s", self._feed
            )
            # Remove all entities.
            count_removed = await self._update_feed_remove_entries(set())
        # Send status update to subscriber.
        await self._status_update(status, count_created, count_updated, count_removed)

    async def _store_feed_entries(
        self, status: str, feed_entries: list[FeedEntry] | None
    ):
        """Keep a copy of all feed entries for future lookups."""
        if feed_entries or status == UPDATE_OK_NO_DATA:
            if status == UPDATE_OK:
                self.feed_entries = {entry.external_id: entry for entry in feed_entries}
        else:
            self.feed_entries.clear()

    async def _update_feed_create_entries(self, feed_external_ids: set[str]) -> int:
        """Create entities after feed update."""
        create_external_ids = feed_external_ids.difference(self._managed_external_ids)
        count_created: int = len(create_external_ids)
        await self._generate_new_entities(create_external_ids)
        return count_created

    async def _update_feed_update_entries(self, feed_external_ids: set[str]) -> int:
        """Update entities after feed update."""
        update_external_ids = self._managed_external_ids.intersection(feed_external_ids)
        count_updated: int = len(update_external_ids)
        await self._update_entities(update_external_ids)
        return count_updated

    async def _update_feed_remove_entries(self, feed_external_ids: set[str]) -> int:
        """Remove entities after feed update."""
        remove_external_ids = self._managed_external_ids.difference(feed_external_ids)
        count_removed: int = len(remove_external_ids)
        await self._remove_entities(remove_external_ids)
        return count_removed

    async def _generate_new_entities(self, external_ids: set[str]):
        """Generate new entities for events."""
        for external_id in external_ids:
            await self._generate_async_callback(external_id)
            _LOGGER.debug("New entity added %s", external_id)
            self._managed_external_ids.add(external_id)

    async def _update_entities(self, external_ids: set[str]):
        """Update entities."""
        for external_id in external_ids:
            _LOGGER.debug("Existing entity found %s", external_id)
            await self._update_async_callback(external_id)

    async def _remove_entities(self, external_ids: set[str]):
        """Remove entities."""
        for external_id in external_ids:
            _LOGGER.debug("Entity not current anymore %s", external_id)
            self._managed_external_ids.remove(external_id)
            await self._remove_async_callback(external_id)

    async def _status_update(
        self, status: str, count_created: int, count_updated: int, count_removed: int
    ):
        """Provide status update."""
        if self._status_async_callback:
            await self._status_async_callback(
                StatusUpdate(
                    status,
                    self.last_update,
                    self.last_update_successful,
                    self.last_timestamp,
                    len(self.feed_entries),
                    count_created,
                    count_updated,
                    count_removed,
                )
            )

    @property
    def last_timestamp(self) -> datetime | None:
        """Return the last timestamp extracted from this feed."""
        return self._feed.last_timestamp

    @property
    def last_update(self) -> datetime | None:
        """Return the last update of this feed."""
        return self._last_update

    @property
    def last_update_successful(self) -> datetime | None:
        """Return the last successful update of this feed."""
        return self._last_update_successful



"""Event Parameters."""

from __future__ import annotations

import logging

from ..consts import XML_TAG_EVENT
from .element import Element
from .event import Event

_LOGGER = logging.getLogger(__name__)


class EventParameters(Element):
    """Represents event parameters."""

    @property
    def events(self) -> list[Event]:
        """Return the events of this feed."""
        items: dict | None = self.attribute([XML_TAG_EVENT])
        entries: list = []
        if items and isinstance(items, list):
            entries = [Event(item) for item in items]
        else:
            # A single item in the feed is not represented as an array.
            entries.append(Event(items))
        return entries



"""XML Parser."""

from __future__ import annotations

from datetime import datetime
import logging

import dateparser
import xmltodict

from ..consts import (
    XML_TAG_CREATIONINFO,
    XML_TAG_CREATIONTIME,
    XML_TAG_DEPTH,
    XML_TAG_EVENT,
    XML_TAG_EVENTPARAMETERS,
    XML_TAG_LATITUDE,
    XML_TAG_LONGITUDE,
    XML_TAG_MAG,
    XML_TAG_MAGNITUDE,
    XML_TAG_ORIGIN,
    XML_TAG_Q_QUAKEML,
    XML_TAG_STATIONCOUNT,
    XML_TAG_TIME,
    XML_TAG_VALUE,
)
from .event_parameters import EventParameters

_LOGGER = logging.getLogger(__name__)

DEFAULT_NAMESPACES = {
    "http://quakeml.org/xmlns/bed/1.2": None,
    "http://quakeml.org/xmlns/quakeml/1.2": "q",
}
KEYS_CHAINS_DATE: list[list[str]] = [
    [
        XML_TAG_Q_QUAKEML,
        XML_TAG_EVENTPARAMETERS,
        XML_TAG_EVENT,
        XML_TAG_ORIGIN,
        XML_TAG_TIME,
        XML_TAG_VALUE,
    ],
    [
        XML_TAG_Q_QUAKEML,
        XML_TAG_EVENTPARAMETERS,
        XML_TAG_EVENT,
        XML_TAG_CREATIONINFO,
        XML_TAG_CREATIONTIME,
    ],
]
KEYS_CHAINS_INT: list[list[str]] = [
    [
        XML_TAG_Q_QUAKEML,
        XML_TAG_EVENTPARAMETERS,
        XML_TAG_EVENT,
        XML_TAG_MAGNITUDE,
        XML_TAG_STATIONCOUNT,
    ],
]
KEY_CHAINS_FLOAT: list[list[str]] = [
    [
        XML_TAG_Q_QUAKEML,
        XML_TAG_EVENTPARAMETERS,
        XML_TAG_EVENT,
        XML_TAG_ORIGIN,
        XML_TAG_LATITUDE,
        XML_TAG_VALUE,
    ],
    [
        XML_TAG_Q_QUAKEML,
        XML_TAG_EVENTPARAMETERS,
        XML_TAG_EVENT,
        XML_TAG_ORIGIN,
        XML_TAG_LONGITUDE,
        XML_TAG_VALUE,
    ],
    [
        XML_TAG_Q_QUAKEML,
        XML_TAG_EVENTPARAMETERS,
        XML_TAG_EVENT,
        XML_TAG_MAGNITUDE,
        XML_TAG_MAG,
        XML_TAG_VALUE,
    ],
    [
        XML_TAG_Q_QUAKEML,
        XML_TAG_EVENTPARAMETERS,
        XML_TAG_EVENT,
        XML_TAG_ORIGIN,
        XML_TAG_DEPTH,
        XML_TAG_VALUE,
    ],
]


class XmlParser:
    """Built-in XML parser."""

    def __init__(self, additional_namespaces: dict | None = None):
        """Initialise the XML parser."""
        self._namespaces: dict = DEFAULT_NAMESPACES
        if additional_namespaces:
            self._namespaces.update(additional_namespaces)

    @staticmethod
    def postprocessor(
        path: list[str], key: str, value: str
    ) -> tuple[str, str | float | int | datetime]:
        """Conduct type conversion for selected keys."""
        try:
            if XmlParser._is_path_in(path, KEY_CHAINS_FLOAT):
                return key, float(value)
            if XmlParser._is_path_in(path, KEYS_CHAINS_INT):
                return key, int(value)
            if XmlParser._is_path_in(path, KEYS_CHAINS_DATE):
                return key, dateparser.parse(
                    value,
                    settings={"TIMEZONE": "UTC", "RETURN_AS_TIMEZONE_AWARE": True},
                )
        except (ValueError, TypeError) as error:
            _LOGGER.warning("Unable to process (%s/%s): %s", key, value, error)
        return key, value

    @staticmethod
    def _is_path_in(path: list[str], chains: list[list[str]]) -> bool:
        """Test if the path is in any of the provided chains."""
        if path and chains:
            new_path: list = [element[0] for element in path]
            for chain in chains:
                if chain == new_path:
                    return True
        return False

    def parse(self, xml: str) -> EventParameters | None:
        """Parse the provided xml."""
        if xml:
            parsed_dict: dict = xmltodict.parse(
                xml,
                process_namespaces=True,
                namespaces=self._namespaces,
                postprocessor=XmlParser.postprocessor,
            )
            if XML_TAG_Q_QUAKEML in parsed_dict:
                return XmlParser._create_feed_from_quakeml(parsed_dict)
        return None

    @staticmethod
    def _create_feed_from_quakeml(parsed_dict: dict) -> EventParameters | None:
        """Create feed from provided RSS data."""
        quakeml: dict = parsed_dict.get(XML_TAG_Q_QUAKEML)
        if XML_TAG_EVENTPARAMETERS in quakeml:
            events: dict = quakeml.get(XML_TAG_EVENTPARAMETERS)
            return EventParameters(events)
        _LOGGER.warning(
            "Invalid structure: Missing top level element %s", XML_TAG_Q_QUAKEML
        )
        return None



"""Description."""

from __future__ import annotations

from ..consts import XML_TAG_TEXT
from .element import Element


class Description(Element):
    """Event description."""

    @property
    def text(self) -> str | None:
        """Return description's text."""
        return self.attribute_with_text([XML_TAG_TEXT])



"""Origin."""

from __future__ import annotations

from datetime import datetime

from ..consts import (
    XML_TAG_DEPTH,
    XML_TAG_DEPTHTYPE,
    XML_TAG_EVALUATIONMODE,
    XML_TAG_EVALUATIONSTATUS,
    XML_TAG_LATITUDE,
    XML_TAG_LONGITUDE,
    XML_TAG_TIME,
    XML_TAG_VALUE,
)
from .element import Element


class Origin(Element):
    """Focal time and geographical location of an earthquake hypocenter."""

    @property
    def latitude(self) -> float | None:
        """Return the hypocenter latitude."""
        latitude: dict | None = self.attribute([XML_TAG_LATITUDE])
        if latitude:
            return latitude.get(XML_TAG_VALUE)
        return None

    @property
    def longitude(self) -> float | None:
        """Return the hypocenter longitude."""
        longitude: dict | None = self.attribute([XML_TAG_LONGITUDE])
        if longitude:
            return longitude.get(XML_TAG_VALUE)
        return None

    @property
    def depth(self) -> float | None:
        """Return depth of hypocenter with respect to the nominal sea level."""
        depth: dict | None = self.attribute([XML_TAG_DEPTH])
        if depth:
            return depth.get(XML_TAG_VALUE)
        return None

    @property
    def depth_type(self) -> str | None:
        """Return type of depth determination."""
        return self.attribute([XML_TAG_DEPTHTYPE])

    @property
    def time(self) -> datetime | None:
        """Return focal time."""
        time: dict | None = self.attribute([XML_TAG_TIME])
        if time:
            return time.get(XML_TAG_VALUE)
        return None

    @property
    def evaluation_mode(self) -> str | None:
        """Return mode of evaluation ."""
        return self.attribute_with_text([XML_TAG_EVALUATIONMODE])

    @property
    def evaluation_status(self) -> str | None:
        """Return status of evaluation ."""
        return self.attribute_with_text([XML_TAG_EVALUATIONSTATUS])



"""Setup of aio_quakeml_client library."""

from setuptools import find_packages, setup

from aio_quakeml_client.__version__ import __version__

NAME = "aio_quakeml_client"
AUTHOR = "Malte Franken"
AUTHOR_EMAIL = "coding@subspace.de"
DESCRIPTION = "An async QuakeML client library."
URL = "https://github.com/exxamalte/python-aio-quakeml-client"

REQUIRES = [
    "aiohttp>=3.7.4,<4",
    "haversine>=2.8.1",
    "xmltodict>=0.13.0",
    "dateparser>=1.2.0",
]


with open("README.md") as fh:
    long_description = fh.read()

setup(
    name=NAME,
    version=__version__,
    author=AUTHOR,
    author_email=AUTHOR_EMAIL,
    description=DESCRIPTION,
    license="Apache-2.0",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url=URL,
    packages=find_packages(exclude=("tests",)),
    classifiers=[
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    install_requires=REQUIRES,
)



"""Constants for feeds and feed entries."""

from typing import Final

ATTR_ATTRIBUTION: Final = "attribution"

CUSTOM_ATTRIBUTE: Final = "custom_attribute"

DEFAULT_REQUEST_TIMEOUT: Final = 10

UPDATE_OK: Final = "OK"
UPDATE_OK_NO_DATA: Final = "OK_NO_DATA"
UPDATE_ERROR: Final = "ERROR"

XML_ATTR_PUBLICID: Final = "@publicID"

XML_CDATA: Final = "#text"

XML_TAG_Q_QUAKEML: Final = "q:quakeml"
XML_TAG_AGENCYID: Final = "agencyID"
XML_TAG_AUTHOR: Final = "author"
XML_TAG_CREATIONINFO: Final = "creationInfo"
XML_TAG_CREATIONTIME: Final = "creationTime"
XML_TAG_DEPTH: Final = "depth"
XML_TAG_DEPTHTYPE: Final = "depthType"
XML_TAG_DESCRIPTION: Final = "description"
XML_TAG_EVENT: Final = "event"
XML_TAG_EVALUATIONMODE: Final = "evaluationMode"
XML_TAG_EVALUATIONSTATUS: Final = "evaluationStatus"
XML_TAG_EVENTPARAMETERS: Final = "eventParameters"
XML_TAG_LATITUDE: Final = "latitude"
XML_TAG_LONGITUDE: Final = "longitude"
XML_TAG_MAG: Final = "mag"
XML_TAG_MAGNITUDE: Final = "magnitude"
XML_TAG_ORIGIN: Final = "origin"
XML_TAG_STATIONCOUNT: Final = "stationCount"
XML_TAG_TEXT: Final = "text"
XML_TAG_TIME: Final = "time"
XML_TAG_TYPE: Final = "type"
XML_TAG_VALUE: Final = "value"



"""Creation info."""

from __future__ import annotations

from ..consts import XML_TAG_AGENCYID, XML_TAG_AUTHOR, XML_TAG_CREATIONTIME
from .element import Element


class CreationInfo(Element):
    """Creation metadata."""

    @property
    def agency_id(self) -> str | None:
        """Return designation of agency that published a resource."""
        return self.attribute_with_text([XML_TAG_AGENCYID])

    @property
    def author(self) -> str | None:
        """Return name describing the author of a resource."""
        return self.attribute_with_text([XML_TAG_AUTHOR])

    @property
    def creation_time(self) -> str | None:
        """Return time of creation of a resource."""
        return self.attribute_with_text([XML_TAG_CREATIONTIME])



"""Base class for any QuakeML elements."""

from __future__ import annotations

import logging

from ..consts import XML_ATTR_PUBLICID, XML_CDATA, XML_TAG_TYPE

_LOGGER = logging.getLogger(__name__)


class Element:
    """Element."""

    def __init__(self, source: dict):
        """Initialise feed."""
        self._source = source

    def __repr__(self):
        """Return string representation of this feed item."""
        return f"<{self.__class__.__name__}({self.public_id})>"

    def attribute(self, names: list[str]) -> str | dict | None:
        """Get an attribute from this element."""
        if self._source and names:
            # Try each name, and return the first value that is not None.
            for name in names:
                value = self._source.get(name, None)
                if value:
                    return value
        return None

    def attribute_with_text(self, names: list[str]) -> str | int:
        """Get an attribute with text from this element."""
        value = self.attribute(names)
        if value and isinstance(value, dict) and XML_CDATA in value:
            # <tag attr="/some.uri">Value</tag>
            value = value.get(XML_CDATA)
        return value

    @staticmethod
    def attribute_in_structure(obj, keys: list[str]) -> str:
        """Return the attribute found under the chain of keys."""
        key: str = keys.pop(0)
        if key in obj:
            return Element.attribute_in_structure(obj[key], keys) if keys else obj[key]
        return ""

    @property
    def public_id(self) -> str | None:
        """Return the public id of this element."""
        return self.attribute([XML_ATTR_PUBLICID])

    @property
    def type(self) -> str | None:
        """Return element's type."""
        return self.attribute_with_text([XML_TAG_TYPE])
